package com.tzld.piaoquan.recommend.server.service.recall;

import com.tzld.piaoquan.recommend.server.common.ThreadPoolFactory;
import com.tzld.piaoquan.recommend.server.common.enums.AppTypeEnum;
import com.tzld.piaoquan.recommend.server.model.Video;
import com.tzld.piaoquan.recommend.server.service.flowpool.FlowPoolConstants;
import com.tzld.piaoquan.recommend.server.service.recall.strategy.*;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.lang3.StringUtils;
import org.springframework.beans.BeansException;
import org.springframework.context.ApplicationContext;
import org.springframework.context.ApplicationContextAware;
import org.springframework.stereotype.Service;

import javax.annotation.PostConstruct;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.concurrent.CountDownLatch;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Future;
import java.util.concurrent.TimeUnit;

/**
 * @author dyp
 */
@Service
@Slf4j
public class RecallService implements ApplicationContextAware {

    private final Map<String, RecallStrategy> strategyMap = new HashMap<>();
    private ApplicationContext applicationContext;
    private ExecutorService pool = ThreadPoolFactory.recallPool();

    @PostConstruct
    public void init() {
        Map<String, RecallStrategy> type = applicationContext.getBeansOfType(RecallStrategy.class);
        for (Map.Entry<String, RecallStrategy> entry : type.entrySet()) {
            RecallStrategy value = entry.getValue();
            strategyMap.put(value.getClass().getSimpleName(), value);
        }
    }

    public RecallResult recall(RecallParam param) {
        List<RecallStrategy> strategies = getRecallStrategy(param);
        // log.info("strategies {}", JSONUtils.toJson(CommonCollectionUtils.toList(strategies, o -> o.pushFrom())));
        CountDownLatch cdl = new CountDownLatch(strategies.size());
        List<Future<RecallResult.RecallData>> recallResultFutures = new ArrayList<>();
        for (final RecallStrategy strategy : strategies) {
            Future<RecallResult.RecallData> future = pool.submit(() -> {
                List<Video> result = strategy.recall(param);
                cdl.countDown();
                return new RecallResult.RecallData(strategy.pushFrom(), result);
            });
            recallResultFutures.add(future);
        }
        try {
            cdl.await(3000, TimeUnit.MILLISECONDS);
        } catch (InterruptedException e) {
            log.error("recall error", e);
            return null;
        }

        List<RecallResult.RecallData> results = new ArrayList<>();
        for (Future<RecallResult.RecallData> f : recallResultFutures) {
            try {
                RecallResult.RecallData data = f.get();
                results.add(data);
            } catch (Exception e) {
                log.error("future get error ", e);
            }
        }

        return new RecallResult(results);
    }

    private List<RecallStrategy> getRecallStrategy(RecallParam param) {
        List<RecallStrategy> strategies = new ArrayList<>();
        if (param.isSpecialRecommend()) {
            strategies.add(strategyMap.get(SpecialRecallStrategy.class.getSimpleName()));
            return strategies;
        }

        //1：通过“产品”控制“召回子策略”
        if (param.getAppType() == AppTypeEnum.LAO_HAO_KAN_VIDEO.getCode()
                || param.getAppType() == AppTypeEnum.ZUI_JING_QI.getCode()) {
            strategies.addAll(getRegionRecallStrategy(param));
        } else {
            //2:通过“流量池标记”控制“流量池召回子策略”
            if (param.getFlowPoolAbtestGroup().equals(FlowPoolConstants.EXPERIMENTAL_FLOW_SET_LEVEL)) {
                strategies.addAll(getRegionRecallStrategy(param));
                strategies.add(strategyMap.get(QuickFlowPoolWithLevelRecallStrategy.class.getSimpleName()));
                strategies.add(strategyMap.get(FlowPoolWithLevelRecallStrategy.class.getSimpleName()));
            } else if (param.getFlowPoolAbtestGroup().equals(FlowPoolConstants.EXPERIMENTAL_FLOW_SET_LEVEL_SCORE)) {
                strategies.addAll(getRegionRecallStrategy(param));
                strategies.add(strategyMap.get(QuickFlowPoolWithLevelScoreRecallStrategy.class.getSimpleName()));
                strategies.add(strategyMap.get(FlowPoolWithLevelScoreRecallStrategy.class.getSimpleName()));
            } else {
                strategies.addAll(getRegionRecallStrategy(param));
                strategies.add(strategyMap.get(QuickFlowPoolWithScoreRecallStrategy.class.getSimpleName()));
                strategies.add(strategyMap.get(FlowPoolWithScoreRecallStrategy.class.getSimpleName()));
            }
        }

        //3:通过“abcode”控制“召回子策略”
        String abCode = param.getAbCode();
        if (abCode == null) {
            // todo 做兜底吗？
            ;
        } else {
            switch (abCode) {
                case "60096":
                    strategies.add(strategyMap.get(RegionHWithoutDupRecallStrategy.class.getSimpleName()));
                    strategies.add(strategyMap.get(SimHotVideoRecallStrategy.class.getSimpleName()));
                    strategies.add(strategyMap.get(ReturnVideoRecallStrategy.class.getSimpleName()));
                    break;
                case "60107":
                case "60106":
                case "60068":
                case "60092":
                case "60094":
                case "60097":
                case "60098":
                case "60101": // 排序str实验
                case "60102": // 通过更改param中的配置实现使用不同数据源 data66 rule68 + 无排序模块
                case "60103": // 增加地域1小时扩量，通过配置实现
                case "60105": // 通过更改param中的配置实现使用不同数据源 data66 rule68 + 有排序模块
                    strategies.add(strategyMap.get(SimHotVideoRecallStrategy.class.getSimpleName()));
                    strategies.add(strategyMap.get(ReturnVideoRecallStrategy.class.getSimpleName()));
                    break;
                case "60104": // 去掉sim的对比实验
                    strategies.add(strategyMap.get(ReturnVideoRecallStrategy.class.getSimpleName()));
                    break;
                case "60110": // 新内容的召回（流量池的Top内容）
                    strategies.add(strategyMap.get(SimHotVideoRecallStrategy.class.getSimpleName()));
                    strategies.add(strategyMap.get(ReturnVideoRecallStrategy.class.getSimpleName()));
                    strategies.add(strategyMap.get(FlowPoolLastDayTopRecallStrategy.class.getSimpleName()));
                    break;
                default:
                    // todo 做兜底吗？
                    break;
            }
        }

        return strategies;
    }

    private List<RecallStrategy> getRegionRecallStrategy(RecallParam param) {

        List<RecallStrategy> strategies = new ArrayList<>();
        if (StringUtils.isNotBlank(param.getRegionCode())) {
            strategies.add(strategyMap.get(RegionHRecallStrategy.class.getSimpleName()));
            strategies.add(strategyMap.get(Region24HRecallStrategy.class.getSimpleName()));
        }
        strategies.add(strategyMap.get(RegionHDupRecallStrategy.class.getSimpleName()));
        strategies.add(strategyMap.get(RegionRelative24HRecallStrategy.class.getSimpleName()));
        strategies.add(strategyMap.get(RegionRelative24HDupRecallStrategy.class.getSimpleName()));

        return strategies;
    }

    @Override
    public void setApplicationContext(ApplicationContext applicationContext) throws BeansException {
        this.applicationContext = applicationContext;
    }
}
